@@ -51,9 +51,25 @@ void leon_init(ambapp_core *core, int level)
 static int irqmp_initied = 0;
 static int eirqs[16];
 
+static void irqmp_add_eirq(struct irqmp_regs *cregs)
+{
+	int eirq;
+	int i;
+
+	eirq = (BYPASS_LOAD_PA(&cregs->mpstatus) >> 16) & 0xf;
+	for (i=0; eirq && i < 16; i++) {
+		if (!eirqs[i]) {
+			eirqs[i] = eirq;
+			break;
+		} else if (eirqs[i] == eirq) {
+			break;
+		}
+	}
+}
+
 int irqmp_dev(ambapp_dev *dev)
 {
-	int i, c, icsel, eirq;
+	int c, icsel, nctrl;
 	ambapp_apbdev *apbslv;
 	struct irqmp_regs *regs, *cregs;
 
@@ -66,21 +82,18 @@ int irqmp_dev(ambapp_dev *dev)
 	}
 
 	/* Record eirqs, so that grgpio can avoid collisions */
-	for (c = 0; c < 16; c++) {
-		/* Logic lifted from leon_kernel.c in linux kernel */
-		icsel = BYPASS_LOAD_PA(&regs->icsel[c/8]);
-		icsel = (icsel >> ((7 - (c & 0x7)) * 4)) & 0xf;
-		cregs = regs + icsel;
+	nctrl = BYPASS_LOAD_PA(&regs->ampctrl) >> 28;
+	if (nctrl) {
+		for (c = 0; c < 16; c++) {
+			/* Logic lifted from leon_kernel.c in linux kernel */
+			icsel = BYPASS_LOAD_PA(&regs->icsel[c/8]);
+			icsel = (icsel >> ((7 - (c & 0x7)) * 4)) & 0xf;
+			cregs = regs + icsel;
 
-		eirq = (BYPASS_LOAD_PA(&cregs->mpstatus) >> 16) & 0xf;
-		for (i=0; eirq && i < 32; i++) {
-			if (!eirqs[i]) {
-				eirqs[i] = eirq;
-				break;
-			} else if (eirqs[i] == eirq) {
-				break;
-			}
+			irqmp_add_eirq(cregs);
 		}
+	} else {
+		irqmp_add_eirq(regs);
 	}
 	return 0;
 }
@@ -583,6 +596,7 @@ void greth_init(ambapp_core *core, int level)
 }
 
 #define GR712RC_SYSTEMID 0x7120e70
+#define GR740_SYSTEMID 0x07401038
 
 void canmux_init(ambapp_core *core, int level)
 {
@@ -906,6 +920,28 @@ struct node *ahb2ahb_create_bus_node(struct ambapp_bus *ainfo, unsigned long ioa
 	return n;
 }
 
+void ahb2ahb_bus_create_iommu_fixup(
+	struct ambapp_bus *ainfo, struct ahb_bus *parent, ambapp_ahbdev *bridge,
+	struct ambapp_pnp_ahb *ahb, int index)
+{
+	unsigned long systemid =
+		BYPASS_LOAD_PA(CONFIG_AMBA_IO_AREA | AMBA_CONF_AREA | 0xff0);
+
+	if ((systemid == GR740_SYSTEMID) && (index <= 9)) {
+		/* GR740 reports always version 0 for all 10 masters.
+		 *  1 x PCI Master
+		 *  1 x PCI DMA Master
+		 *  4 x SpW masters verion: 1
+		 */
+		if (index == 0)
+			*(unsigned int *)&ahb->id |= 1 << 5; /* PCI master */
+		else if (index == 1)
+			*(unsigned int *)&ahb->id |= 1 << 5; /* PCI DMA master */
+		else if (index >= 4 && index <= 7)
+			*(unsigned int *)&ahb->id |= 1 << 5; /* SpaceWire masters */
+	}
+}
+
 /* AHB-AHB IOMMU Creation */
 void ahb2ahb_bus_create_iommu(struct ambapp_bus *ainfo, struct ahb_bus *parent,
 			ambapp_ahbdev *bridge)
@@ -951,9 +987,13 @@ void ahb2ahb_bus_create_iommu(struct ambapp_bus *ainfo, struct ahb_bus *parent,
 			if (mst) {
 				struct ambapp_pnp_ahb ahb_buf;
 				_memset((void*)&ahb_buf, 0, sizeof(ahb_buf));
-				*(int *)&(ahb_buf.id) = mst;
+				*(int *)&(ahb_buf.id) = mst & 0xffffffe0;
 				*(int *)&(ahb_buf.mbar[0]) = 0; /* mst */
-				
+
+				/* perform fixups */
+				ahb2ahb_bus_create_iommu_fixup(
+					ainfo, parent, bridge, &ahb_buf, j);
+
 				/* Build a AHB Master Device */
 				pdev->dev_type = DEV_AHB_MST;
 				ambapp_dev_init_ahb(subbus, &ahb_buf, pdev);
@@ -1484,11 +1524,12 @@ ambapp_core *ambapp_core_create(
 					 */
 					ahb->irq[0] |= 0x80;
 					core->ahbslv = ahb;
-					break;
+					goto ahbscandone;
 				}
 			}
 		}
 	}
+ahbscandone:
 
 	/* Find a APB Slave for Device */
 	if (options & 0x4) {
@@ -1505,11 +1546,12 @@ ambapp_core *ambapp_core_create(
 					 */
 					apb->irq[0] |= 0x80;
 					core->apbslv = apb;
-					break;
+					goto apbscandone;
 				}
 			}
 		}
 	}
+apbscandone:
 
 	core->index = ambapp_core_cnt++;
 	return core;
